MODULE systeme_lineaire

  !------------------------------------------------------------------------
  ! Goal: Solve linear systems with different methods
  !------------------------------------------------------------------------

  USE mes_types
  USE param
  USE functions

  IMPLICIT NONE

CONTAINS

  SUBROUTINE syslin(mat,vec,sol,x0,siz,precis,opt,precond,num)

    INTEGER,INTENT(inout) :: opt
    INTEGER,INTENT(in) :: siz,precond
    REAL,INTENT(IN) :: precis
    TYPE (diag),DIMENSION(:),POINTER :: mat
    REAL,DIMENSION(:),POINTER :: vec,x0
    REAL,DIMENSION(:),POINTER :: sol
    INTEGER,INTENT(in),DIMENSION(:) ::num
    !print*,'Entree dans syslin'  !,siz,size(vec),size(sol),size(mat(1)%tab)

    IF ((precond==1).OR.(precond==3)) CALL precond_diag(mat,vec,x0,siz)
    IF ((opt==2).AND.((precond==2).OR.(precond==3))) opt=4

    SELECT CASE (opt)

    CASE (1) 
       CALL gradient_conjugue(mat,vec,sol,x0,siz,precis)
    CASE (2)
       CALL bi_gradient(mat,vec,sol,x0,siz,precis)
    CASE (3)
       CALL gmres(mat,vec,sol,x0,siz,precis)
    CASE (4)
       CALL bigrad_lu_3(mat,vec,sol,x0,siz,precis,num)
    CASE default
       PRINT*,'Linear system: method not programmed:',opt
       STOP
    END SELECT

  END SUBROUTINE syslin

  SUBROUTINE gradient_conjugue(mat,vec,sol,x0,siz,precis)

    !----------------------------------------------------------------------
    ! CONJUGATE GRADIENT
    !----------------------------------------------------------------------

    INTEGER,INTENT(in) :: siz
    REAL,INTENT(IN) :: precis
    TYPE (diag),DIMENSION(:),POINTER :: mat
    REAL,DIMENSION(:),POINTER :: vec,x0
    REAL,DIMENSION(:),POINTER :: sol

    INTEGER      :: i
    REAL   :: alfa,beta,residu,newresidu,tester
    REAL, DIMENSION(siz)    :: temp, res, desc  

    i = 1 
    sol = x0 
    desc = vec - mat_dot_vec(mat,sol,siz)
    res = desc  
    residu = scalar(res,res)
    tester=SQRT(residu)
    IF (tester < precis) RETURN

    ! Begin iterating
    !----------------
    DO
       temp = mat_dot_vec(mat,desc,siz)
       alfa = residu/scalar(temp,desc)
       sol = sol + (alfa * desc)
       res = res - (alfa * temp)
       newresidu =scalar(res,res)
       beta = newresidu/residu
       desc = res + (beta * desc)
       residu = newresidu
       tester=SQRT(residu/scalar(sol,sol))
       i = i + 1

       ! Exit
       !-----
       IF( tester < precis ) THEN
          EXIT
       END IF
       IF( i > siz+10) THEN
          PRINT *, 'Divergence CG -> iter =', i, 'residual =', residu
          STOP
       END IF
    END DO

  END SUBROUTINE gradient_conjugue

  SUBROUTINE bi_gradient(mat,vec,sol,x0,siz,precis)

    !----------------------------------------------------------------------
    ! CONJUGATE BI-GRADIENT
    !----------------------------------------------------------------------

    INTEGER,INTENT(in) :: siz
    REAL,INTENT(IN) :: precis
    TYPE(diag),DIMENSION(:),POINTER :: mat
    REAL,DIMENSION(:),POINTER :: vec,x0
    REAL,DIMENSION(:),POINTER :: sol

    INTEGER      :: i,j,k,l
    REAL   :: alfa,beta,residu,residu_bis,norm,tester
    REAL, DIMENSION(siz)    :: temp, desc1,res2,desc2,res1

    norm=SQRT(scalar_pointer(vec,vec))

    tester=norm
    i = 1 
    sol = x0 

    desc1 = vec - mat_dot_vec_pointer(mat,sol,siz)
    desc2 = vec - mat_dot_vec_trans_pointer(mat,sol,siz) 
    res1 = desc1
    res2 = desc2
    residu_bis = scalar(res1,res2)
    residu = scalar(res1,res1)
    tester=SQRT(residu/norm)
    IF (tester < precis) RETURN

    ! Begin iterating
    !----------------
    DO
       temp = mat_dot_vec(mat,desc1,siz)
       alfa = residu_bis/scalar(temp,desc2)
       sol = sol + (alfa * desc1)
       res1 = res1 - (alfa * temp)
       res2 = res2 - (alfa * mat_dot_vec_trans(mat,desc2,siz))
       beta = scalar(res1,res2)/residu_bis
       desc1 = res1+(beta * desc1)
       desc2 = res2 + (beta * desc2)
       residu_bis=scalar(res1,res2)
       residu =scalar(res1,res1)
       tester=SQRT(residu/scalar_pointer(sol,sol))

       ! Exit
       !-----
       IF( tester < precis ) EXIT
       IF( i > siz+10) THEN
          PRINT *, 'Divergence BI-CG -> iter ='&
               &, i, 'residual =', residu
          STOP
       END IF

       i = i + 1

    END DO

  END SUBROUTINE bi_gradient


  SUBROUTINE gmres(mat,vec,sol,x0,siz,precis)

    !----------------------------------------------------------------------
    ! GMRES
    !----------------------------------------------------------------------

    INTEGER,INTENT(in) :: siz
    REAL,INTENT(IN) :: precis
    TYPE (diag),DIMENSION(:),POINTER :: mat
    REAL,DIMENSION(:),POINTER :: vec,x0
    REAL,DIMENSION(:),POINTER :: sol

    INTEGER      :: i
    REAL   :: alfa,beta,sav,tester
    REAL, DIMENSION(siz)    :: temp, res, desc1,desc2

    i=0
    sol=x0
    res=vec-mat_dot_vec(mat,sol,siz)
    desc1=res
    desc2=mat_dot_vec(mat,desc1,siz)

    ! Begin iterating
    !----------------
    DO
       tester=SQRT(scalar(res,res)/scalar(sol,sol))
       IF (tester<precis) EXIT
       IF (i>siz+10) THEN
          PRINT*,'Divergence GMRES i=',i,'residual=',SQRT(scalar(res,res))
          CALL desallocation
          STOP
       END IF
       i=i+1

       sav=scalar(desc2,desc2)
       alfa=scalar(res,desc2)/sav
       sol=sol+alfa*desc1
       res=res-alfa*desc2
       temp=mat_dot_vec(mat,res,siz)
       beta=scalar(temp,desc2)/sav
       desc1=res+beta*desc1
       desc2=temp+beta*desc2
    END DO

  END SUBROUTINE gmres

  SUBROUTINE precond_diag(mat,vec,x0,siz)

    !----------------------------------------------------------------------
    ! DIAGONAL PRE-CONDITIONING
    !----------------------------------------------------------------------

    INTEGER,INTENT(in) :: siz
    TYPE (diag),DIMENSION(:),POINTER :: mat
    REAL,DIMENSION(:),POINTER :: vec,x0

    INTEGER :: i,j
    REAL :: val

    DO i=1,siz
       val=1.0/mat(1)%tab(i)
       DO j=1,SIZE(mat)
          mat(j)%tab(i)=mat(j)%tab(i)*val
       END DO
       vec(i)=vec(i)*val
       x0(i)=x0(i)*val
    END DO

  END SUBROUTINE precond_diag

  SUBROUTINE bigrad_lu_3 (mat,vec,sol,x0,siz,precis,num)

    !----------------------------------------------------------------------
    ! CONJUGATE BI-GRADIENT WITH LU PRE-CONDITIONING
    !----------------------------------------------------------------------

    INTEGER,INTENT(in) :: siz
    REAL,INTENT(IN) :: precis
    TYPE (diag),DIMENSION(:),POINTER :: mat
    REAL,DIMENSION(:),POINTER :: vec,x0
    REAL,DIMENSION(:),POINTER :: sol
    INTEGER,INTENT(in),DIMENSION(7) :: num

    REAL,DIMENSION(2,siz) :: TI,SU

    INTEGER      :: i,j
    REAL   :: alfa,beta,sav1,sav2,residu,omega
    REAL, DIMENSION(siz)    :: temp1,temp2,res1,res2, desc1,desc2,H1,H2

    H1=0.0
    H2=0.0
    TI=0.0
    SU=0.0
    residu=0.0
    sav1=0.0
    sav2=0.0
    i=0

    CALL inc_gauss_3(mat,TI,SU,siz,num)

    sol=x0

    res1 = vec-mat_dot_vec_pointer(mat,sol,siz)
    res2 = vec-mat_dot_vec_trans_pointer(mat,sol,siz)

    CALL trisup_3 (TI,SU,H1,res1,mat(num(3))%dist,mat(num(4))%dist,siz)

    CALL trisup_trans_3 (TI,SU,H2,res2,mat(num(3))%dist,mat(num(4))%dist,siz)

    desc1=H1
    desc2=H2
    sav1=scalar(H1,res2)

    DO
       residu=MAXVAL(ABS(res1))
       IF (residu<precis) EXIT
       i=i+1
       temp1 = mat_dot_vec(mat,desc1,siz)
       alfa = sav1/scalar(desc2,temp1)
       sol = sol+alfa*desc1
       res1 = res1-alfa*temp1
       temp2 = mat_dot_vec_trans(mat,desc2,siz)
       res2 = res2-alfa*temp2

       CALL trisup_3 (TI,SU,H1,res1,mat(num(3))%dist,mat(num(4))%dist,siz)
       CALL trisup_trans_3 (TI,SU,H2,res2,mat(num(3))%dist,mat(num(4))%dist,siz)

       sav2 = scalar(H1,res2)
       beta = sav2/sav1
       sav1 = sav2
       desc1 = H1+beta*desc1
       desc2 = H2+beta*desc2

    END DO

  END SUBROUTINE bigrad_lu_3


  SUBROUTINE inc_gauss_3 (mat,TI,SU,siz,num)

    !-------------------------------------------------------------------
    ! Create matrices L and U
    !-------------------------------------------------------------------

    INTEGER,INTENT(in) :: siz
    TYPE(diag),DIMENSION(:),POINTER :: mat
    REAL,DIMENSION(:,:),INTENT(inout) :: TI,SU
    INTEGER,INTENT(in),DIMENSION(7) :: num

    INTEGER :: i,mp,ms

    mp=mat(num(2))%dist
    
    TI=0.0
    SU=0.0

    TI(1,1:siz) = 1.0
    SU(2,1:siz-1) = mat(num(2))%tab(1:siz-1)
    SU(1,1) = mat(num(1))%tab(1)

    DO i=2,siz
       TI(2,i) = mat(num(3))%tab(i)/SU(1,i-1)
       SU(1,i) = mat(num(1))%tab(i)-TI(2,i)*SU(2,i-1)
    END DO


  END SUBROUTINE inc_gauss_3


  SUBROUTINE trisup_3 (TI,SU,H1,R1,mp,ms,siz)

    !--------------------------------------------------------------------
    ! Goal: solve upper and lower triangular linear systems corresponding
    !       with LU decomposition
    !-------------------------------------------------------------------

    INTEGER,INTENT(in) :: mp,ms,siz
    REAL,DIMENSION(:,:),INTENT(inout) :: TI,SU
    REAL,DIMENSION(:),INTENT(inout) :: R1
    REAL,DIMENSION(:),INTENT(inout) :: H1

    INTEGER :: i
    REAL,DIMENSION(siz) :: X1


    ! Compute de X1 such that TI*X1 = R1
    !--------------------------------
    X1(1) = R1(1)/TI(1,1)
    DO i=2,siz
       X1(i) = (R1(i)-TI(2,i)*X1(i-1))/TI(1,i)
    END DO

    ! Compute H1 such that SU*H1 = X1 and then mat*H1 = TI*SU*H1 = R1
    !----------------------------------------------------------------
    H1(siz) = X1(siz)/SU(1,siz)
    DO i=siz-1,1,-1
       H1(i) = (X1(i)-SU(2,i)*H1(i+1))/SU(1,i)
    END DO

  END SUBROUTINE trisup_3


  SUBROUTINE trisup_trans_3 (TI,SU,H1,R1,mp,ms,siz)

    !-----------------------------------------------------------------------
    ! Goal: solve the transpose of lower and upper triangular linear systems
    !       corresponding with LU decomposition
    !-----------------------------------------------------------------------

    INTEGER,INTENT(in) :: mp,ms,siz
    REAL,DIMENSION(:,:),INTENT(in) :: TI,SU
    REAL,DIMENSION(:),INTENT(in) :: R1
    REAL,DIMENSION(:),INTENT(inout) :: H1

    INTEGER :: i
    REAL,DIMENSION(siz) :: X1

    ! Compute de X1 such that tSU*X1 = R1
    !------------------------------------
    X1(1) = R1(1)/SU(1,1)
    DO i=2,siz
       X1(i) = (R1(i)-SU(2,i-1)*X1(i-1))/SU(1,i)
    END DO

    ! Compute H1 such that tTI*H1 = X1 and then tmat*H1 = tSU*tTI*H1 = R1
    !--------------------------------------------------------------------
    H1(siz) = X1(siz)/TI(1,siz)
    DO i=siz-1,1,-1
       H1(i) = (X1(i)-TI(2,i+1)*H1(i+1))/TI(1,i)
    END DO
    
  END SUBROUTINE trisup_trans_3




END MODULE systeme_lineaire
