module solver_1d

  USE param
  USE mes_types
  USE functions
  USE systeme_lineaire

CONTAINS

  SUBROUTINE transport_1d (SB,Vx_1d,dt)

    !---------------------------------------------------------------------
    ! Goal: Iterate the pure drift scheme
    !---------------------------------------------------------------------

    IMPLICIT NONE

    REAL,INTENT(in) :: dt
    REAL,DIMENSION(:),POINTER :: SB,Vx_1d
    REAL,DIMENSION(mx_1d) :: sigmaxp,sigmaxm
    REAL,DIMENSION(mx_1d) :: Lp,Lm

    INTEGER :: k,i,j
    REAL :: Vp,Vm,Wp,Wm,ps,Vx_1d1,Vx_1d2

    INTEGER :: opt

    opt=1

    SELECT CASE(schema_transport)

    CASE(1) 
       !---------------------------------------------------------------------
       ! Godunov scheme
       !---------------------------------------------------------------------

       !---------------------------------------------------------------------
       ! Numerical fluxes
       !--------------------
       ! Interior interfaces
       !--------------------
       DO i=1,mx_1d-1
          Vp=MAX(0.,Vx_1d(i))
          Vm=MIN(0.,Vx_1d(i))
          flux_1d(i)=Vp*SB(i)+Vm*SB(i+1)
       END DO
       flux_1d(mx_1d)=0.0
       flux_1d(0)=0.0
       !---------------------------------------------------------------------

       ! Update the numerical solution at time tn+1
       !-------------------------------------------
       DO i=1,mx_1d
          SB9_1d(i)=SB(i)-dt/deltax_1d(i)*(flux_1d(i)-flux_1d(i-1))
       END DO

       SB(:) = SB9_1d(1:mx_1d)

    CASE(2) 
       !---------------------------------------------------------------------
       ! Godunov scheme with limiter
       !---------------------------------------------------------------------

       ! Coefficients sigma
       !--------------------
       DO i=1,mx_1d-1
          Vp=MAX(0.,Vx_1d(i))
          Vm=MIN(0.,Vx_1d(i))
          sigmaxp(i)=Vp*(1.0)!-Vp*dt/deltax_1d(i))
          sigmaxm(i)=Vm*(1.0)!+Vm*dt/deltax_1d(i+1))
       END DO

       ! Limiters
       !---------
       DO i=2,Mx_1d-1
          Wp = SB(i+1)-SB(i)
          Wm=SB(i)-SB(i-1)
          Vx_1d1=MAX(0.,Vx_1d(i-1))
          Vx_1d2=MAX(0.,Vx_1d(i))
          Lp(i) = limiter(Vx_1d1*Wm,Vx_1d2*Wp,dxp_1d(i)/deltax_1d(i),opt)
          Vx_1d1=MIN(0.,Vx_1d(i-1))
          Vx_1d2=MIN(0.,Vx_1d(i))
          Lm(i) = limiter(Vx_1d2*Wp,Vx_1d1*Wm,dxp_1d(i-1)/deltax_1d(i),opt)
       END DO
       Lp(1)=0.   ; Lp(Mx_1d) = 0.
       Lm(1)=0.   ; Lm(Mx_1d) = 0.

       !----------------------------------------------------------------------
       ! Numerical flux_1des
       !----------------------------------------------------------------------
       ! Interior interfaces
       !--------------------
       DO i=1,mx_1d-1
          Vp=MAX(0.,Vx_1d(i))
          Vm=MIN(0.,Vx_1d(i))
          Wp=(SB(i+1)-SB(i))/dxp_1d(i)
          flux_1d(i)=Vp*SB(i)+Vm*SB(i+1)
          flux_1d(i)=flux_1d(i)+.5&
               &*Wp*(Lp(i)*sigmaxp(i)*deltax_1d(i)&
               &-Lm(i+1)*sigmaxm(i)*deltax_1d(i+1))
       END DO

       flux_1d(0)=0.0
       flux_1d(mx_1d)=0.0
       !----------------------------------------------------------------------

       ! Update the solution at time tn+1
       !----------------------------------
       DO i=1,mx_1d
          SB9_1d(i)=SB(i)-dt/deltax_1d(i)*(flux_1d(i)-flux_1d(i-1))
       END DO

       SB(:) = SB9_1d(1:mx_1d)

    END SELECT


  END SUBROUTINE transport_1d

  SUBROUTINE calc_diff_1d (mat,sm,Vx_1d,dt)

    !-----------------------------------------------------------------------
    ! Goal: Assemble the three diagonals of diffusion matrix 
    !       and the right-hand side
    !-----------------------------------------------------------------------

    IMPLICIT NONE

    REAL,INTENT(IN) :: dt

    REAL,DIMENSION(:),POINTER :: sm,Vx_1d
    TYPE(diag),DIMENSION(:),POINTER :: mat

    INTEGER :: i,j,k,m
    REAL :: dr,dl,dc,ci,cio,d11,Vp,Vm


    DO i=1,SIZE(mat)
       mat(i)%tab=0.0
    END DO
    sm=0.0

    !-------------------------------------------------------------------
    !                          INTERIOR  MESHES                        !
    !-------------------------------------------------------------------
    DO i=2,mx_1d-1
       m=i
       dr=diff11_1d(i,diff_1d)/dxp_1d(i)
       dl=diff11_1d(i-1,diff_1d)/dxp_1d(i-1)
       mat(1)%tab(m)=mat(1)%tab(m)+dr+dl
       mat(2)%tab(m)=mat(2)%tab(m)-dr
       mat(3)%tab(m)=mat(3)%tab(m)-dl
    END DO


    !------------------------------------------------------------------
    !                           LEFT BOUNDARY                         !
    !------------------------------------------------------------------
    i=1
    m=i  
    dr=diff11_1d(i,diff_1d)/dxp_1d(i)
    dl=diff11_1d(i-1,diff_1d)/(dxp_1d(i-1)/2.0)
    mat(1)%tab(m)=mat(1)%tab(m)+dr
    mat(2)%tab(m)=mat(2)%tab(m)-dr

    !------------------------------------------------------------------
    !                           RIGHT BOUNDARY                        !
    !------------------------------------------------------------------
    i=mx_1d
    m=i
    dr=diff11_1d(i,diff_1d)/(deltax_1d(i)/2.0)
    dl=diff11_1d(i-1,diff_1d)/dxp_1d(i-1)
    mat(1)%tab(m)=mat(1)%tab(m)+dl
    mat(3)%tab(m)=mat(3)%tab(m)-dl
    mat(1)%tab(m)=mat(1)%tab(m)+dr
    sm(m)=sm(m)+dr*0.0

    ! Drift
    IF (schema_transport==3) then
       DO i=1,mx_1d-1
          Vp=MAX(0.,Vx_1d(i))
          Vm=MIN(0.,Vx_1d(i))
          
          mat(1)%tab(i)=mat(1)%tab(i)+Vp
          mat(3)%tab(i+1)=mat(3)%tab(i+1)-Vp
          mat(2)%tab(i)=mat(2)%tab(i)+Vm
          mat(1)%tab(i+1)=mat(1)%tab(i+1)-Vm
       END DO
    END IF

  END SUBROUTINE calc_diff_1d

  SUBROUTINE diffusion_1d (mat,sm,sub,tau,dt)

    !---------------------------------------------------------------------
    ! Goal: Iterate the pure diffusion scheme
    !---------------------------------------------------------------------

    IMPLICIT NONE

    REAL,INTENT(in) :: dt,tau

    TYPE(diag),DIMENSION(:),POINTER :: mat
    REAL,DIMENSION(:),POINTER :: sub
    REAL,DIMENSION(:),POINTER :: sm

    INTEGER :: i,j,k,l,m
    REAL :: ci,cio,d11,val,val2,T_bar

    ! Complete matrix and right-hand side
    !------------------------------------
    DO i=1,mx_1d
       m=i
       ! Matrix
       !--------
       mat(1)%tab(m)=mat(1)%tab(m)+deltax_1d(i)/dt
       ! Right-hand side
       !----------------
       sm(m)=sm(m)+deltax_1d(i)*sub(i)/dt
    END DO


    ! Coefficient that is not in any of the three diagonals
    !------------------------------------------------------
    !weight_1d=1.0!MAX(0.0,dt-C%tau)/dt

    
!!$IF (iter==919) then
!!$    DO i=1,mx_1d-1
!!$    print*,mat(3)%tab(i+1),mat(1)%tab(i),mat(2)%tab(i)
!!$ end DO
!!$print*,'sm=',sm
!!$print*,dt
!!$end IF
    ! Solve the linear system
    !-------------------------    
    CALL syslin (mat,sm,sub,sub,mx_1d,test_cv_diff,&
         &method_syslin_diff,diff_precond,(/1,2,3,8,11,13,17/))

    ! Calculate the probability mass that crosses the threshold
    !-----------------------------------------------------------
!!$    IF ((C%tau>1d-15).AND.(weight>1d-15)) THEN
!!$       mass_out=mass_out-weight*dt*diff_1d(mx_1d)&
!!$            &*sol_1D(mx_1d)/(deltax_1d(mx_1d)/2.0)
!!$    END IF

  END SUBROUTINE diffusion_1d

  

END module  solver_1d
